==> py_agent_tools-0.0.4.tar.gz: py_agent_tools-0.0.4/agent_tools/credential_pool_base.py <==
import asyncio
from abc import ABC
from datetime import datetime
from enum import Enum
from functools import cached_property
from typing import Any, Callable, Coroutine, Protocol, Union
from urllib.parse import urlparse

from pydantic import BaseModel

from agent_tools._log import log
from agent_tools.settings import AccountCredential


class StatusType(str, Enum):
    ACTIVE = "active"
    INACTIVE = "inactive"
    ERROR = "error"
    RATE_LIMITED = "rate_limited"


class ModelCredential(BaseModel):
    id: str
    base_url: str = ""
    api_key: str = ""
    account_info: dict = {}
    api_version: str = ""
    model_name: str = ""
    deployment: str = ""


class ModelCredentialWithStatus(BaseModel):
    id: str
    credential: ModelCredential
    status: StatusType
    last_checked: datetime


class ModelCredentialStatistics(BaseModel):
    model_name: str
    active: int = 0
    inactive: int = 0
    error: int = 0
    rate_limited: int = 0


def make_id(
    host_name: str,
    model_name: str,
    deployment: str,
) -> str:
    id = ''
    if host_name:
        id += host_name
    if model_name:
        id += f'_{model_name}'
    if deployment:
        id += f'_{deployment}'
    if not id:
        raise ValueError('id is empty')
    return id


# fmt: off
class CredentialPoolProtocol(Protocol):
    async def get_best(self) -> ModelCredential:
        ...

    async def update_status(self, credential: ModelCredential, status: StatusType) -> None:
        ...
# fmt: on


class RemoteCredentialPool(CredentialPoolProtocol):
    def __init__(self, url: str):
        self.url = url

    async def get_best(self) -> ModelCredential:
        raise NotImplementedError()

    async def update_status(self, credential: ModelCredential, status: StatusType) -> None:
        raise NotImplementedError()


class CredentialPoolBase(ABC, CredentialPoolProtocol):
    """
    Credential pool is a pool of credentials for a specific model.
    It is used to:
        - select the best credential for a specific model.
        - get the statistics of the credential pool.
        - update the status of the credential.
    """

    def __init__(
        self,
        account_credentials: list[AccountCredential],
        target_model: str,
        validate_fn: Callable[[ModelCredential], Union[bool, Coroutine[Any, Any, bool]]],
    ):
        self.account_credentials = account_credentials
        self.target_model = target_model
        self._running = False
        self._lock = asyncio.Lock()
        self._health_check_task: asyncio.Task | None = None
        self.credential_pool: dict[str, ModelCredentialWithStatus] = {}
        self.validate_fn = validate_fn
        self.health_check_interval = 60

    @cached_property
    def model_credentials(self) -> list[ModelCredential]:
        model_credentials: list[ModelCredential] = []
        for account_credential in self.account_credentials:
            for model in account_credential.supported_models:
                if model.model_name != self.target_model:
                    continue
                model_credentials.append(
                    ModelCredential(
                        id=make_id(
                            host_name=urlparse(account_credential.base_url).netloc,
                            model_name='',
                            deployment=model.deployment,
                        ),
                        base_url=account_credential.base_url,
                        api_key=account_credential.api_key,
                        account_info=account_credential.account_info,
                        api_version=model.api_version,
                        model_name=model.model_name,
                        deployment=model.deployment,
                    )
                )
        return model_credentials

    async def get_best(self) -> ModelCredential:
        return min(
            [
                credential.credential
                for credential in self.credential_pool.values()
                if credential.status == StatusType.ACTIVE
            ],
            key=lambda x: self.credential_pool[x.id].last_checked,
        )

    async def update_status(
        self,
        credential: ModelCredential,
        status: StatusType,
    ):
        async with self._lock:
            if credential.id in self.credential_pool.keys():
                self.credential_pool[credential.id].status = status
                self.credential_pool[credential.id].last_checked = datetime.now()
            else:
                self.credential_pool[credential.id] = ModelCredentialWithStatus(
                    id=credential.id,
                    credential=credential,
                    status=status,
                    last_checked=datetime.now(),
                )

    async def _validate_and_update_status(self, credential: ModelCredential):
        result = self.validate_fn(credential)
        if asyncio.iscoroutine(result):
            result = await result
        if result:
            await self.update_status(credential, StatusType.ACTIVE)
        else:
            await self.update_status(credential, StatusType.INACTIVE)

    async def _log_status(self):
        log.info('当前凭证池状态:')
        for credential in self.credential_pool.values():
            log.info(
                f'{credential.id}: {credential.status}: '
                f'{credential.last_checked.strftime("%Y-%m-%d %H:%M:%S")}'
            )

    async def _health_check(self):
        for credential in self.credential_pool.values():
            # if credential.status == StatusType.ACTIVE:
            #     continue
            await self._validate_and_update_status(credential.credential)
        await self._log_status()

    async def _health_check_loop(self):
        log.info('执行首次健康检查')
        for credential in self.model_credentials:
            await self._validate_and_update_status(credential)
        log.info('首次健康检查完成')
        await self._log_status()

        log.info(f'健康检查循环开始, 间隔 {self.health_check_interval} 秒')
        while self._running:
            await asyncio.sleep(self.health_check_interval)
            await self._health_check()

    async def _initialize(self):
        for credential in self.model_credentials:
            await self.update_status(credential, StatusType.ACTIVE)
        log.info('凭证池初始化完成')
        await self._log_status()

    def _cleanup(self):
        self.credential_pool = {}
        log.info('凭证池清理完成')

    async def start(self):
        self._running = True
        await self._initialize()
        log.info('启动后台健康检查任务')
        self._health_check_task = asyncio.create_task(self._health_check_loop())

    def stop(self):
        self._running = False
        if self._health_check_task is not None:
            self._health_check_task.cancel()
            self._health_check_task = None
        self._cleanup()

==> py_agent_tools-0.0.5.tar.gz: py_agent_tools-0.0.5/agent_tools/credential_pool_base.py <==
import asyncio
from abc import ABC
from datetime import datetime
from enum import Enum
from functools import cached_property
from typing import Any, Callable, Coroutine, Protocol, Union

from pydantic import BaseModel

from agent_tools._log import log
from agent_tools.settings import AccountCredential


class StatusType(str, Enum):
    ACTIVE = "active"
    INACTIVE = "inactive"
    ERROR = "error"
    RATE_LIMITED = "rate_limited"


class ModelCredential(BaseModel):
    id: str
    base_url: str = ""
    api_key: str = ""
    account_info: dict = {}
    api_version: str = ""
    model_name: str = ""
    deployment: str = ""


class ModelCredentialWithStatus(BaseModel):
    id: str
    credential: ModelCredential
    status: StatusType
    last_checked: datetime


class ModelCredentialStatistics(BaseModel):
    model_name: str
    active: int = 0
    inactive: int = 0
    error: int = 0
    rate_limited: int = 0


def make_id(
    account_id: str,
    group_id: str,
    model_name: str,
    deployment: str,
) -> str:
    id = ''
    if account_id:
        id += account_id
    if group_id:
        id += f'_{group_id}'
    if model_name:
        id += f'_{model_name}'
    if deployment:
        id += f'_{deployment}'
    if not id:
        raise ValueError('id is empty')
    return id


# fmt: off
class CredentialPoolProtocol(Protocol):
    async def get_best(self) -> ModelCredential:
        ...

    async def update_status(self, credential: ModelCredential, status: StatusType) -> None:
        ...
# fmt: on


class RemoteCredentialPool(CredentialPoolProtocol):
    def __init__(self, url: str):
        self.url = url

    async def get_best(self) -> ModelCredential:
        raise NotImplementedError()

    async def update_status(self, credential: ModelCredential, status: StatusType) -> None:
        raise NotImplementedError()


class CredentialPoolBase(ABC, CredentialPoolProtocol):
    """
    Credential pool is a pool of credentials for a specific model.
    It is used to:
        - select the best credential for a specific model.
        - get the statistics of the credential pool.
        - update the status of the credential.
    """

    def __init__(
        self,
        account_credentials: list[AccountCredential],
        target_model: str,
        validate_fn: Callable[[ModelCredential], Union[bool, Coroutine[Any, Any, bool]]],
        health_check_interval: int = 600,
    ):
        self.account_credentials = account_credentials
        self.target_model = target_model
        self._running = False
        self._lock = asyncio.Lock()
        self._health_check_task: asyncio.Task | None = None
        self.credential_pool: dict[str, ModelCredentialWithStatus] = {}
        self.validate_fn = validate_fn
        self.health_check_interval = health_check_interval

    @cached_property
    def model_credentials(self) -> list[ModelCredential]:
        model_credentials: list[ModelCredential] = []
        for account_credential in self.account_credentials:
            for model in account_credential.supported_models:
                if model.model_name != self.target_model:
                    continue
                model_credentials.append(
                    ModelCredential(
                        id=make_id(
                            account_id=account_credential.account_id,
                            group_id=account_credential.group_id,
                            model_name=model.model_name,
                            deployment=model.deployment,
                        ),
                        base_url=account_credential.base_url,
                        api_key=account_credential.api_key,
                        account_info=account_credential.account_info,
                        api_version=model.api_version,
                        model_name=model.model_name,
                        deployment=model.deployment,
                    )
                )
        if len(model_credentials) == 0:
            raise ValueError(f'No model credentials found for {self.target_model}')
        return model_credentials

    async def get_best(self) -> ModelCredential:
        return min(
            [
                credential.credential
                for credential in self.credential_pool.values()
                if credential.status == StatusType.ACTIVE
            ],
            key=lambda x: self.credential_pool[x.id].last_checked,
        )

    async def update_status(
        self,
        credential: ModelCredential,
        status: StatusType,
    ):
        async with self._lock:
            if credential.id in self.credential_pool.keys():
                self.credential_pool[credential.id].status = status
                self.credential_pool[credential.id].last_checked = datetime.now()
            else:
                self.credential_pool[credential.id] = ModelCredentialWithStatus(
                    id=credential.id,
                    credential=credential,
                    status=status,
                    last_checked=datetime.now(),
                )

    async def _validate_and_update_status(self, credential: ModelCredential):
        result = self.validate_fn(credential)
        if asyncio.iscoroutine(result):
            result = await result
        if result:
            await self.update_status(credential, StatusType.ACTIVE)
        else:
            await self.update_status(credential, StatusType.INACTIVE)

    async def _log_status(self):
        log.info('当前凭证池状态:')
        for credential in self.credential_pool.values():
            log.info(
                f'{credential.id}: {credential.status}: '
                f'{credential.last_checked.strftime("%Y-%m-%d %H:%M:%S")}'
            )

    async def _health_check(self):
        for credential in self.credential_pool.values():
            # if credential.status == StatusType.ACTIVE:
            #     continue
            await self._validate_and_update_status(credential.credential)
        await self._log_status()

    async def _health_check_loop(self):
        log.info('执行首次健康检查')
        for credential in self.model_credentials:
            await self._validate_and_update_status(credential)
        log.info('首次健康检查完成')
        await self._log_status()

        log.info(f'健康检查循环开始, 间隔 {self.health_check_interval} 秒')
        while self._running:
            await asyncio.sleep(self.health_check_interval)
            await self._health_check()

    async def _initialize(self):
        for credential in self.model_credentials:
            await self.update_status(credential, StatusType.ACTIVE)
        log.info('凭证池初始化完成')
        await self._log_status()

    def _cleanup(self):
        self.credential_pool = {}
        log.info('凭证池清理完成')

    async def start(self):
        self._running = True
        await self._initialize()
        log.info('启动后台健康检查任务')
        self._health_check_task = asyncio.create_task(self._health_check_loop())

    def stop(self):
        self._running = False
        if self._health_check_task is not None:
            self._health_check_task.cancel()
            self._health_check_task = None
        self._cleanup()
